import 'dart:async';
import 'dart:isolate';
import 'dart:ui';

import 'package:android_alarm_manager_plus/android_alarm_manager_plus.dart';
import 'package:codenic_alarm/codenic_alarm.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:shared_preferences/shared_preferences.dart';

/// {@template CodenicAlarmAndroid}
/// A Flutter plugin to schedule alarms for Android.
/// {@endtemplate}
class CodenicAlarmAndroid extends CodenicAlarm {
  /// The name of the [SendPort] used by this plugin to send the currently
  /// active [AlarmInfo] to its corresponding [ReceivePort] from the
  /// [CodenicAlarmOverlayHandler].
  static const alarmInfoSendPortName = 'codenic_alarm/alarm_info_send_port';

  /// The [MethodChannel] used to invoke commands to the platform side of this
  /// plugin.
  @visibleForTesting
  MethodChannel get methodChannel => const MethodChannel('codenic_alarm');

  /// The [BasicMessageChannel] used to receive events from the platform side
  /// of this plugin.
  BasicMessageChannel<Object?> get messageChannel => const BasicMessageChannel(
        'codenic_alarm/message_channel',
        JSONMessageCodec(),
      );

  /// The [SharedPreferences] instance used to store the registered alarm IDs.
  ///
  /// This is initialized in the [initialize] method.
  late final SharedPreferences _preferences;

  /// The key used to store the registered alarm IDs in the [_preferences].
  final _registeredAlarmsPreferences = 'registered_alarms';

  /// Indicates whether the [initialize] methods has been called.
  bool _isInitialized = false;

  @override
  FutureOr<bool> initialize() async {
    if (_isInitialized) {
      return true;
    }

    _isInitialized = true;

    await Future.wait([
      AndroidAlarmManager.initialize(),
      Future(() async => _preferences = await SharedPreferences.getInstance()),
    ]);

    _initializeMessageChannelHandler();

    return true;
  }

  /// Initializes the [messageChannel] handler by forwarding the received
  /// messages from the platform channel to the [CodenicAlarmOverlayHandler].
  void _initializeMessageChannelHandler() =>
      messageChannel.setMessageHandler((message) async {
        switch (message) {
          case {'event': 'alarm_started'}:
            IsolateNameServer.lookupPortByName(
              CodenicAlarmAndroid.alarmInfoSendPortName,
            )?.send(message['alarm']);

          case {'event': 'alarm_dismissed'}:
            IsolateNameServer.lookupPortByName(
              CodenicAlarmAndroid.alarmInfoSendPortName,
            )?.send(null);
        }

        return message;
      });

  @override
  FutureOr<bool> schedule({
    required Duration delay,
    required AlarmInfo alarmInfo,
  }) =>
      scheduleAt(schedule: DateTime.now().add(delay), alarmInfo: alarmInfo);

  @override
  FutureOr<bool> scheduleAt({
    required DateTime schedule,
    required AlarmInfo alarmInfo,
  }) async {
    final registeredAlarms =
        (_preferences.getStringList(_registeredAlarmsPreferences) ?? []).toSet()
          ..add(alarmInfo.id.toString());

    await _preferences.setStringList(
      _registeredAlarmsPreferences,
      registeredAlarms.toList(),
    );

    return AndroidAlarmManager.oneShotAt(
      schedule,
      alarmInfo.id,
      _alarmCallback,
      allowWhileIdle: true,
      alarmClock: true,
      exact: true,
      rescheduleOnReboot: true,
      wakeup: true,
      params: alarmInfo.toJson(),
    );
  }

  @override
  FutureOr<void> dismissAlarm() =>
      CodenicAlarmAndroid().methodChannel.invokeMethod('dismissAlarm');

  @override
  FutureOr<void> cancel({required int id}) => AndroidAlarmManager.cancel(id);

  @override
  FutureOr<void> cancelAll() async {
    final registeredAlarms =
        _preferences.getStringList(_registeredAlarmsPreferences);

    if (registeredAlarms == null) {
      return;
    }

    await Future.wait([
      for (final alarm in registeredAlarms)
        AndroidAlarmManager.cancel(int.parse(alarm)),
    ]);

    await _preferences.remove(_registeredAlarmsPreferences);
  }

  @override
  FutureOr<AlarmInfo?> activeAlarm() async {
    final alarmData =
        await methodChannel.invokeMapMethod<String, dynamic>('getActiveAlarm');

    if (alarmData == null) {
      return null;
    }

    return AlarmInfo.fromJson(alarmData);
  }
}

/// The callback function invoked to play an alarm.
@pragma('vm:entry-point')
Future<void> _alarmCallback(int id, Map<String, dynamic> data) async {
  final codenicAlarm = CodenicAlarmAndroid();

  await codenicAlarm.initialize();
  await codenicAlarm.methodChannel.invokeMethod('startAlarm', data);

  IsolateNameServer.lookupPortByName(CodenicAlarmAndroid.alarmInfoSendPortName)
      ?.send(data);
}
