import 'dart:isolate';
import 'dart:ui';

import 'package:codenic_alarm/src/codenic_alarm.dart';
import 'package:codenic_alarm/src/models/alarm_info.dart';
import 'package:codenic_alarm/src/platforms/codenic_alarm_android.dart';
import 'package:flutter/material.dart';

/// {@template CodenicAlarmOverlayHandler}
/// A widget that renders a [Widget] on top of the app when an alarm is
/// triggered.
/// {@endtemplate}
class CodenicAlarmOverlayHandler extends StatefulWidget {
  /// {@macro CodenicAlarmOverlayHandler}
  const CodenicAlarmOverlayHandler({
    required this.child,
    required this.alarmOverlayBuilder,
    this.alarmOverlayLayoutBuilder,
    super.key,
  });

  /// The [Widget] that is rendered below the alarm overlay.
  final Widget child;

  /// The builder that builds the alarm overlay.
  ///
  /// Make sure that this widget contains a dismiss button that calls
  /// [CodenicAlarm.dismissAlarm] to dismiss the active alarm and remove the
  /// overlay.
  final Widget Function(BuildContext context, AlarmInfo alarmInfo)
      alarmOverlayBuilder;

  /// The builder that builds the layout of the [child] and the widget
  /// generated by the [alarmOverlayBuilder].
  ///
  /// If there is no active alarm, the `alarmOverlay` is `null`.
  final Widget Function(
    BuildContext context,
    Widget child,
    Widget? alarmOverlay,
  )? alarmOverlayLayoutBuilder;

  @override
  State<CodenicAlarmOverlayHandler> createState() =>
      _CodenicAlarmOverlayHandlerState();
}

class _CodenicAlarmOverlayHandlerState
    extends State<CodenicAlarmOverlayHandler> {
  /// A [ReceivePort] that receives the currently active [AlarmInfo] from the
  /// [CodenicAlarm] plugin.
  static final _alarmInfoReceivePort = ReceivePort();

  static final _alarmInfoReceivePortStream =
      _alarmInfoReceivePort.asBroadcastStream();

  bool _isInitialAlarmInfoLoaded = false;

  AlarmInfo? _initialAlarmInfo;

  @override
  void initState() {
    super.initState();

    _registerCodenicAlarmSendPort();
    _loadInitialAlarmInfo();
  }

  /// Loads the initial [AlarmInfo] from the [CodenicAlarm] plugin.
  ///
  /// This is useful for fetching the [AlarmInfo] when the app has just been
  /// opened and the alarm is already active.
  Future<void> _loadInitialAlarmInfo() async {
    _initialAlarmInfo = await CodenicAlarm.instance.activeAlarm();
    setState(() => _isInitialAlarmInfoLoaded = true);
  }

  /// Registers the [_alarmInfoReceivePort]'s send port to the [CodenicAlarm]
  /// plugin.
  ///
  /// This enables the currently active [AlarmInfo] from the plugin to be
  /// received by this app.
  void _registerCodenicAlarmSendPort() {
    IsolateNameServer.removePortNameMapping(
      CodenicAlarmAndroid.alarmInfoSendPortName,
    );

    IsolateNameServer.registerPortWithName(
      _alarmInfoReceivePort.sendPort,
      CodenicAlarmAndroid.alarmInfoSendPortName,
    );
  }

  @override
  Widget build(BuildContext context) {
    if (!_isInitialAlarmInfoLoaded) {
      return widget.child;
    }

    return StreamBuilder(
      initialData: _initialAlarmInfo?.toJson(),
      stream: _alarmInfoReceivePortStream,
      builder: (context, snapshot) {
        final alarmInfo = snapshot.hasData
            ? AlarmInfo.fromJson(snapshot.data as Map<String, dynamic>)
            : null;

        return widget.alarmOverlayLayoutBuilder?.call(
              context,
              widget.child,
              alarmInfo != null
                  ? widget.alarmOverlayBuilder(context, alarmInfo)
                  : null,
            ) ??
            Stack(
              children: [
                widget.child,
                AnimatedSwitcher(
                  duration: const Duration(milliseconds: 300),
                  child: alarmInfo != null
                      ? widget.alarmOverlayBuilder(context, alarmInfo)
                      : null,
                ),
              ],
            );
      },
    );
  }
}
